#!/opt/homebrew/Caskroom/miniforge/base/envs/tf_38/bin/python
import numpy as np
import torch
import torchvision
import matplotlib.pyplot as plt

def save_img(file_name, img):
  f="./pic/%s.png"%file_name
  img=np.transpose(img.numpy(), (1, 2, 0))
  plt.imsave(f, img)

# RBM model with torch
class RBM(torch.nn.Module):
  def __init__(self, n_vis=784, n_hid=500, k=5):
    super(RBM, self).__init__()
    # weigths rand initialize
    self.W=torch.nn.Parameter(torch.randn(n_hid, n_vis)*1e-2)
    # bias 0 initialize
    self.v_bias=torch.nn.Parameter(torch.zeros(n_vis))
    self.h_bias=torch.nn.Parameter(torch.zeros(n_hid))
    # Number of Divergence
    self.k=k
  
  def sampling(self, p):
    # from Gaussian distribution: achieve probability
    _p=p-torch.autograd.Variable(torch.rand(p.size()))
    # if _p>p return 1 else return -1;
    p_sign=torch.sign(_p)
    # with Relu-> if _p>p return 1 else return 0;
    return torch.nn.functional.relu(p_sign)
  
  def v_to_h(self, v):
    # P(h_i=1|v)=sigmoid(c_i+W_i*v)
    p_h=torch.sigmoid(torch.nn.functional.linear(v, self.W, self.h_bias))
    sample_h=self.sampling(p_h)
    return p_h, sample_h
  
  def h_to_v(self, h):
    # P(v_j=1|h)=sigmoid(b_j+W'_j*h)
    p_v=torch.sigmoid(torch.nn.functional.linear(h, self.W.t(), self.v_bias))
    sample_v=self.sampling(p_v)
    return p_v, sample_v
  
  def forward(self, v):
    p_h0, h0=self.v_to_h(v)
    _h=h0
    for _ in range(self.k): # Divergence count
      _p_v, _v=self.h_to_v(_h)
      _p_h, _h=self.v_to_h(_v)
    
    return v, _v    # return initial v and after k times divergenced _v
  
  def freeEnergy(self, v):
    # F(v)=-b'v-SUM_i[ log(1+exp(c_i+W_iv)) ]
    vbias=v.mv(self.v_bias) # b'v
    wx_b=torch.nn.functional.linear(v, self.W, self.h_bias) # c+Wv
    sum_i=torch.log(1+torch.exp(wx_b))
    ssum=torch.sum(sum_i, dim=1)
    # F(v)^(1)
    return (-ssum-vbias).mean()

if __name__=="__main__":
  # k=1 conversion 1 times
  rbm=RBM(k=1)
  # optimizer Adam
  optimizer=torch.optim.Adam(rbm.parameters(), 1e-3)
  
  batch_size=64
  train_loader = torch.utils.data.DataLoader(
    torchvision.datasets.MNIST('./', train=True, download=True, transform=torchvision.transforms.Compose([torchvision.transforms.ToTensor()])), batch_size=batch_size)
  
  test=torchvision.datasets.MNIST('./', train=False, transform=torchvision.transforms.Compose([torchvision.transforms.ToTensor()]))
  # test with 32 data
  sample_data=test.data[:32, :].view(-1, 784)
  sample_data=sample_data.type(torch.FloatTensor)/255.
  
  v, v1=rbm(sample_data)
  save_img("label", torchvision.utils.make_grid(v.view(32, 1, 28, 28).data))
  save_img("before_train", torchvision.utils.make_grid(v1.view(32, 1, 28, 28).data))
  
  for epoch in range(100):
    loss=[]
    for _, (data, label) in enumerate(train_loader):
      data=torch.autograd.Variable(data.view(-1, 784))
      # we consider bernoulli h
      _sample_data=data.bernoulli()
      # forward
      v, v1=rbm(_sample_data)
      _loss=rbm.freeEnergy(v)-rbm.freeEnergy(v1)
      loss.append(_loss.data.item())
      optimizer.zero_grad()
      _loss.backward()
      optimizer.step()
    print("EPOCH: %d\r" %(epoch+1), flush=True, end="")
    save_img("epoch%d" %(epoch+1), torchvision.utils.make_grid(v1.view(32, 1, 28, 28).data))
  
  v, v1=rbm(sample_data)
  save_img("after_train", torchvision.utils.make_grid(v1.view(32, 1, 28, 28).data))
  plt.plot(loss, "co")
  plt.show()
